from contextvars import ContextVar
from pathlib import Path

from aiopath import AsyncPath
from pydantic import BaseModel

root_path: AsyncPath = AsyncPath(Path.cwd())


class Action(BaseModel, arbitrary_types_allowed=True):
    name: str
    pkg: str = "acb"
    module: str = ""
    path: AsyncPath = AsyncPath(__file__) / "actions"


action_registry: ContextVar[list[Action]] = ContextVar("action_registry", default=[])


class AdapterNotFound(Exception):
    pass


def register_actions() -> list[Action]:
    actions_path = Path(root_path / "actions")
    if not actions_path.exists():
        actions_path = Path(__file__).parent
    actions = {
        a.stem: a
        for a in actions_path.iterdir()
        if a.is_file() and not a.name.startswith("_")
    }
    _actions = []
    for action_name, path in actions.items():
        action = next((a for a in action_registry.get() if a.name == action_name), None)
        if action:
            del action
        _action = Action(name=action_name, path=AsyncPath(path))
        action_registry.get().append(_action)
        _actions.append(_action)
    return _actions



import typing as t

from hishel import (
    AsyncCacheClient,
    AsyncRedisStorage,
    Controller,
)
from hishel._utils import generate_key
from httpcore import Request
from httpx import Response as HttpxResponse
from redis.asyncio import Redis as AsyncRedis
from acb.depends import depends
from ._base import RequestsBase, RequestsBaseSettings


class RequestsSettings(RequestsBaseSettings): ...


class Requests(RequestsBase):
    storage: AsyncRedisStorage
    controller: Controller

    def cache_key(self, request: Request, body: bytes) -> str:
        key = generate_key(request, body)
        return f"{self.config.app.name}:httpx:{key}"

    async def get(self, url: str, timeout: int = 5) -> HttpxResponse:
        async with AsyncCacheClient(
            storage=self.storage, controller=self.controller
        ) as client:
            return await client.get(url, timeout=timeout)

    async def post(
        self, url: str, data: dict[str, t.Any], timeout: int = 5
    ) -> HttpxResponse:
        async with AsyncCacheClient(
            storage=self.storage, controller=self.controller
        ) as client:
            return await client.post(url, data=data, timeout=timeout)

    async def put(
        self, url: str, data: dict[str, t.Any], timeout: int = 5
    ) -> HttpxResponse:
        async with AsyncCacheClient(
            storage=self.storage, controller=self.controller
        ) as client:
            return await client.put(url, data=data, timeout=timeout)

    async def delete(self, url: str, timeout: int = 5) -> HttpxResponse:
        async with AsyncCacheClient(
            storage=self.storage, controller=self.controller
        ) as client:
            return await client.delete(url, timeout=timeout)

    @depends.inject
    async def init(self) -> None:  # type: ignore
        self.storage = AsyncRedisStorage(
            client=AsyncRedis(
                host=self.config.cache.host.get_secret_value(),
                port=self.config.cache.port,
                # db=self.config.requests.cache_db,
            ),
            ttl=self.config.requests.cache_ttl,
        )
        self.controller = Controller(key_generator=self.cache_key)  # type: ignore


depends.set(Requests)



import asyncio
import os
import typing as t
from abc import ABC, abstractmethod
from contextvars import ContextVar
from enum import Enum
from functools import cached_property
from pathlib import Path
from secrets import token_bytes, token_urlsafe
from string import punctuation

import nest_asyncio
from aiopath import AsyncPath
from inflection import titleize, underscore
from pydantic import BaseModel, SecretStr, field_validator
from pydantic._internal._utils import deep_update
from pydantic.fields import FieldInfo
from pydantic_settings import SettingsConfigDict
from pydantic_settings.sources import SettingsError
from acb import register_pkg
from acb.actions.encode import dump, load
from acb.adapters import (
    adapter_registry,
    import_adapter,
    root_path,
    settings_path,
    tmp_path,
)
from acb.depends import depends

nest_asyncio.apply()

register_pkg()


async def get_version() -> str:
    pyproject_toml = root_path.parent / "pyproject.toml"
    version_file = root_path / "_version"
    if await pyproject_toml.exists():
        _version = (await load.toml(pyproject_toml)).get("project").get("version")
        await version_file.write_text(_version)
        return _version
    elif await version_file.exists():
        return await version_file.read_text()
    return "0.0.1"


project: str = ""
app_name: str = ""
debug: dict[str, bool] = {}
_deployed: bool = os.getenv("DEPLOYED", "False").lower() == "true"
_secrets_path: AsyncPath = tmp_path / "secrets"
_app_secrets: ContextVar[set[str]] = ContextVar("_app_secrets", default=set())


class Platform(str, Enum):
    aws = "aws"
    gcp = "gcp"
    azure = "azure"


async def init_app() -> None:
    for path in (
        tmp_path,
        _secrets_path,
        AsyncPath(settings_path),
    ):
        await path.mkdir(exist_ok=True)


asyncio.run(init_app())


def gen_password(size: int = 10) -> str:
    return token_urlsafe(size)


class PydanticBaseSettingsSource(ABC):
    adapter_name: str = "app"

    def __init__(
        self,
        settings_cls: type["Settings"],
        secrets_dir: AsyncPath = _secrets_path,
    ) -> None:
        self.settings_cls = settings_cls
        self.adapter_name = underscore(
            self.settings_cls.__name__.replace("Settings", "")
        )
        self.secrets_dir = secrets_dir
        self.config = settings_cls.model_config

    def get_model_secrets(self) -> dict[str, FieldInfo]:
        return {
            "_".join((self.adapter_name, n)): v
            for n, v in self.settings_cls.model_fields.items()
            if v.annotation is SecretStr
        }

    @abstractmethod
    async def __call__(self) -> dict[str, t.Any]:
        raise NotImplementedError


class InitSettingsSource(PydanticBaseSettingsSource):
    def __init__(
        self, settings_cls: type["Settings"], init_kwargs: dict[str, t.Any]
    ) -> None:
        self.init_kwargs = init_kwargs
        super().__init__(settings_cls)

    async def __call__(self) -> dict[str, t.Any]:
        return self.init_kwargs

    def __repr__(self) -> str:
        return f"InitSettingsSource(init_kwargs={self.init_kwargs!r})"


class FileSecretSource(PydanticBaseSettingsSource):
    async def get_field_value(self, field_name: str) -> SecretStr | None:
        path = self.secrets_path / field_name
        if await path.is_file():
            return SecretStr((await path.read_text()).strip())

    async def __call__(self) -> dict[str, t.Any]:
        data = {}
        self.secrets_path = await AsyncPath(self.secrets_dir).expanduser()
        if not await self.secrets_path.exists():
            await self.secrets_path.mkdir(parents=True, exist_ok=True)
        model_secrets = self.get_model_secrets()
        for field_key in model_secrets:
            field_key = field_key.removeprefix(f"{app_name}_")
            if field_key in _app_secrets.get():
                continue
            try:
                field_value = await self.get_field_value(field_key)
            except Exception as e:
                raise SettingsError(
                    f'error getting value for field "{field_key}" from '
                    f'source "{self.__class__.__name__}"'
                ) from e
            if field_value is not None:
                field_name = field_key.removeprefix(f"{self.adapter_name}_")
                data[field_name] = field_value
                _app_secrets.get().add(field_key)
        return data

    def __repr__(self) -> str:
        return f"FileSecretSource(_secrets_dir={self.secrets_dir!r})"


class ManagerSecretSource(PydanticBaseSettingsSource):
    @cached_property
    def manager(self):
        manager = depends.get(import_adapter("secret"))
        return manager

    async def load_secrets(self) -> t.Any:
        data = {}
        adapter_secrets = self.get_model_secrets()
        missing_secrets = {
            n: v for n, v in adapter_secrets.items() if n not in _app_secrets.get()
        }
        if missing_secrets and self.manager:
            manager_secrets = await self.manager.list(self.adapter_name)
            for field_key, field_value in missing_secrets.items():
                stored_field_key = "_".join((app_name, field_key))
                secret_path = self.secrets_dir / field_key
                if not await secret_path.exists():
                    if field_key not in manager_secrets:
                        await self.manager.create(
                            stored_field_key, field_value.default.get_secret_value()
                        )
                secret = await self.manager.get(stored_field_key)
                await secret_path.write_text(secret)
                field_name = field_key.removeprefix(f"{self.adapter_name}_")
                data[field_name] = SecretStr(secret)
                _app_secrets.get().add(field_key)
        return data

    async def __call__(self) -> t.Any:
        data = await self.load_secrets()
        return data


class YamlSettingsSource(PydanticBaseSettingsSource):
    async def load_yml_settings(self) -> t.Any:
        global project, app_name, debug
        if self.adapter_name == "secret":
            return {}
        yml_path = AsyncPath(settings_path / f"{self.adapter_name}.yml")
        if not await yml_path.exists() and not _deployed:
            dump_settings = {
                name: info.default
                for name, info in self.settings_cls.model_fields.items()
                if (info.annotation is not SecretStr)
                and ("Optional" not in (str(info.annotation)))
            }
            await dump.yaml(dump_settings, yml_path)
        yml_settings = await load.yaml(yml_path)
        if self.adapter_name == "debug":
            for adapter in [
                a
                for a in adapter_registry.get()
                if a.category not in yml_settings.keys()
            ]:
                yml_settings[adapter.category] = False
            debug = yml_settings
        if not _deployed:
            await dump.yaml(yml_settings, yml_path, sort_keys=True)
        if self.adapter_name == "app":
            project = yml_settings["project"]
            app_name = yml_settings["name"]
        return yml_settings or {}

    async def __call__(self) -> dict[str, t.Any]:
        data = {}
        if Path.cwd().name != "acb":
            for field_name, field in (await self.load_yml_settings()).items():
                if field is not None:
                    data[field_name] = field
            return data
        return data


class Settings(BaseModel):
    loggers: t.Optional[list[str]] = []

    model_config: t.ClassVar[SettingsConfigDict] = SettingsConfigDict(  # type: ignore
        extra="allow",
        arbitrary_types_allowed=True,
        validate_default=True,
        secrets_dir=_secrets_path,
        protected_namespaces=("model_", "settings_"),
    )

    def __init__(
        __pydantic_self__,  # type: ignore
        _secrets_dir: AsyncPath = _secrets_path,
        **values: t.Any,
    ) -> None:
        build_settings = __pydantic_self__._settings_build_values(
            values,
            _secrets_dir=_secrets_dir,
        )
        build_settings = asyncio.run(build_settings)
        super().__init__(**build_settings)

    def __getattr__(self, item: str) -> t.Any:
        return super().__getattr__(item)  # type: ignore

    async def _settings_build_values(
        self,
        init_kwargs: dict[str, t.Any],
        _secrets_dir: AsyncPath = _secrets_path,
    ) -> dict[str, t.Any]:
        secrets_dir = _secrets_dir or self.model_config.get("secrets_dir")
        init_settings = InitSettingsSource(self.__class__, init_kwargs=init_kwargs)
        file_secret_settings = FileSecretSource(
            self.__class__,
            secrets_dir=secrets_dir,
        )
        manager_secret_settings = ManagerSecretSource(
            self.__class__,
            secrets_dir=secrets_dir,
        )
        yaml_settings = YamlSettingsSource(self.__class__)

        sources = self.settings_customize_sources(
            self.__class__,
            init_settings=init_settings,
            yaml_settings=yaml_settings,
            file_secret_settings=file_secret_settings,
            manager_secret_settings=manager_secret_settings,
        )
        return deep_update(*reversed([await source() for source in sources]))

    @classmethod
    def settings_customize_sources(
        cls,  # noqa: F841
        settings_cls: t.Type["Settings"],
        init_settings: PydanticBaseSettingsSource,
        yaml_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
        manager_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            yaml_settings,
            file_secret_settings,
            manager_secret_settings,
        )


class DebugSettings(Settings):
    production: bool = False
    secrets: bool = False
    logger: bool = False


class AppSettings(Settings):
    name: str = "myapp"
    secret_key: SecretStr = SecretStr(token_urlsafe(32))
    secure_salt: SecretStr = SecretStr(str(token_bytes(32)))
    title: t.Optional[str] = None
    domain: t.Optional[str] = None
    platform: t.Optional[Platform] = None
    project: t.Optional[str] = None
    region: t.Optional[str] = None
    timezone: t.Optional[str] = "US/Pacific"
    version: t.Optional[str] = asyncio.run(get_version())

    def model_post_init(self, __context: t.Any) -> None:  # noqa: F841
        self.title = self.title or titleize(self.name)

    @field_validator("name")
    @classmethod
    def cloud_compliant_app_name(cls, v: str) -> str:  # noqa: F841
        not_ok = [" ", "_", "."]
        _name = v
        for p in not_ok:
            _name = _name.replace(p, "-")
        for p in punctuation.replace("-", ""):
            _name = _name.replace(p, "")
        app_name = _name.strip("-").lower()
        if len(app_name) < 3:
            raise SystemExit("App name to short")
        elif len(app_name) > 63:
            raise SystemExit("App name to long")
        return app_name


class Config(BaseModel, extra="allow"):
    deployed: bool = _deployed
    debug: t.Optional[Settings] = None
    app: t.Optional[Settings] = None

    def __getattr__(self, item: str) -> t.Any:
        return super().__getattr__(item)  # type: ignore

    def init(self) -> None:
        self.debug = DebugSettings()
        self.app = AppSettings()


depends.set(Config)
depends.get(Config).init()

Logger = import_adapter()


@depends.inject
def initialize_acb(config: Config = depends(), logger: Logger = depends()) -> None:  # type: ignore
    logger.info(f"App path: {root_path}")
    logger.info(f"App deployed: {config.deployed}")


initialize_acb()


class AdapterBase(ABC):
    config: Config = depends()
    logger: Logger = depends()  # type: ignore

    @abstractmethod
    async def init(self) -> None:
        raise NotImplementedError



from contextvars import ContextVar
from inspect import currentframe
from pathlib import Path

from pydantic import BaseModel
from acb.actions import Action, register_actions
from acb.adapters import Adapter, register_adapters


class Pkg(BaseModel):
    name: str
    actions: list[Action] = []
    adapters: list[Adapter] = []


pkg_registry: ContextVar[list[Pkg]] = ContextVar("pkg_registry", default=[])


def register_pkg() -> None:
    name = Path(currentframe().f_back.f_code.co_filename).parent.stem
    registry = pkg_registry.get()
    if name not in [p.name for p in registry]:
        actions = register_actions()
        adapters = register_adapters()
        pkg = Pkg(name=name, actions=actions, adapters=adapters)
        if pkg.name == "acb":
            registry.insert(0, pkg)
        else:
            registry.append(Pkg(name=name, actions=actions, adapters=adapters))



import typing as t
from functools import cached_property

from fsspec.asyn import AsyncFileSystem
from fsspec.implementations.dirfs import DirFileSystem
from acb.depends import depends
from ._base import StorageBase, StorageBaseSettings


class StorageSettings(StorageBaseSettings): ...


class Storage(StorageBase):
    file_system: t.Any = DirFileSystem

    @cached_property
    def client(self) -> AsyncFileSystem:
        return self.file_system(path=self.config.storage.local_path, asynchronous=True)


depends.set(Storage)



import typing as t

from bevy import dependency, get_repository
from bevy import inject as inject_dependency


class Depends:
    @staticmethod
    def inject(func: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        return inject_dependency(func)

    @staticmethod
    def set(class_: t.Any) -> t.Any:
        return get_repository().set(class_, class_())

    @staticmethod
    def get(class_: t.Any) -> t.Any:
        if isinstance(class_, str):
            from acb.adapters import import_adapter

            class_ = import_adapter(class_)
        return get_repository().get(class_)

    def __call__(self, *args: t.Any, **kwargs: t.Any):
        return dependency()


depends = Depends()



import typing as t

from acb.config import Settings


class LoggerBaseSettings(Settings):
    verbose: bool = False
    deployed_level: str = "WARNING"
    log_level: t.Optional[str] = "INFO"


class LoggerBase: ...



import asyncio
import sys
import typing as t
from contextvars import ContextVar
from importlib import import_module, util
from inspect import currentframe, stack
from pathlib import Path

from aiopath import AsyncPath
from inflection import camelize
from msgspec.yaml import decode as yaml_decode
from pydantic import BaseModel
from acb.actions.encode import yaml_encode
from acb.depends import depends

root_path: AsyncPath = AsyncPath(Path.cwd())
tmp_path: AsyncPath = root_path / "tmp"


class Adapter(BaseModel, arbitrary_types_allowed=True):
    name: str
    class_name: str
    category: str
    pkg: str = "acb"
    module: str = ""
    enabled: bool = False
    installed: bool = False

    path: AsyncPath = AsyncPath(Path(__file__) / "adapters")

    def __str__(self) -> str:
        return self.__repr__()


adapter_registry: ContextVar[list[Adapter]] = ContextVar("adapter_registry", default=[])
_install_lock: ContextVar[list[str]] = ContextVar("install_lock", default=[])

settings_path: Path = Path(root_path / "settings")
app_settings_path: Path = settings_path / "app.yml"
adapter_settings_path: Path = settings_path / "adapters.yml"


class AdapterNotFound(Exception):
    pass


def get_adapter(category: str) -> Adapter | None:
    adapter = next(
        (a for a in adapter_registry.get() if a.category == category and a.enabled),
        None,
    )
    return adapter


def get_enabled_adapters() -> list[Adapter]:
    return [a for a in adapter_registry.get() if a.enabled]


def get_installed_adapters() -> list[Adapter]:
    return [a for a in adapter_registry.get() if a.installed]


async def _import_adapter(adapter_category: str, config: t.Any) -> t.Any:
    try:
        adapter = get_adapter(adapter_category)
    except AttributeError:
        raise AdapterNotFound(
            f"{adapter_category} adapter not found - please make sure one is "
            f" configured in app.yml"
        )
    try:
        module = import_module(adapter.module)
    except ModuleNotFoundError:
        spec = util.spec_from_file_location(adapter.path.stem, adapter.path)
        module = util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(module)
        sys.modules[adapter.name] = module
    adapter_class = getattr(module, adapter.class_name)
    if not adapter.installed and adapter.name not in _install_lock.get():
        _install_lock.get().append(adapter.name)
        adapter_settings_class_name = f"{adapter.class_name}Settings"
        # adapter_class = getattr(module, adapter.class_name)
        adapter_settings_class = getattr(module, adapter_settings_class_name)
        adapter_settings = adapter_settings_class()
        setattr(config, adapter_category, adapter_settings)
        await depends.get(adapter_class).init()
        adapter.installed = True
        _install_lock.get().remove(adapter.name)
        if adapter_category != "logger":
            logger = depends.get(import_adapter("logger"))
            logger.info(f"{adapter.class_name} adapter installed")
    return adapter_class


def import_adapter(adapter_categories: t.Optional[str | list[str]] = None) -> t.Any:
    from acb.config import Config

    config = depends.get(Config)
    if isinstance(adapter_categories, str):
        adapter_categories = [adapter_categories]
    if not adapter_categories:
        adapter_categories = [
            c.strip()
            for c in (
                stack()[1][4][0].split("=")[0].strip().lower()  # type: ignore
            ).split(",")
        ]
    classes = asyncio.run(
        asyncio.gather(  # type: ignore
            *[_import_adapter(c, config) for c in adapter_categories]
        )
    )
    return classes[0] if len(adapter_categories) < 2 else classes


def path_adapters(path: Path) -> dict[str, list[Path]]:
    return {
        a.stem: [
            m for m in a.iterdir() if not m.name.startswith("_") and m.suffix == ".py"
        ]
        for a in path.iterdir()
        if a.is_dir() and not a.name.startswith("__")
    }


def create_adapter(path: Path) -> Adapter:
    return Adapter(
        name=path.stem,
        class_name=camelize(path.parent.stem),
        category=path.parent.stem,
        module=".".join(path.parts[-4:]).removesuffix(".py"),
        pkg=path.parent.parent.parent.stem,
        path=AsyncPath(path),
    )


def extract_adapter_modules(modules: list[Path], adapter: str) -> list[Adapter]:
    adapter_modules = [create_adapter(p) for p in modules]
    return [a for a in adapter_modules if a.category == adapter]


def register_adapters() -> list[Adapter]:
    adapters_path = (
        Path(currentframe().f_back.f_back.f_code.co_filename).parent / "adapters"
    )
    pkg_adapters = path_adapters(adapters_path)
    adapters: list[Adapter] = []
    for adapter_name, modules in pkg_adapters.items():
        _modules = extract_adapter_modules(modules, adapter_name)
        adapters.extend(_modules)
    if not adapter_settings_path.exists():
        settings_path.mkdir(exist_ok=True)
        print(adapter_registry.get())
        categories = {a.category for a in set(adapter_registry.get() + adapters)}
        adapter_settings_path.write_bytes(
            yaml_encode(
                {cat: None for cat in categories},
                sort_keys=True,
            )
        )
    _adapters = yaml_decode(adapter_settings_path.read_text())
    _adapters.update(
        {a.category: None for a in adapters if a.category not in _adapters}
    )
    adapter_settings_path.write_bytes(yaml_encode(_adapters, sort_keys=True))
    enabled_adapters = {a: m for a, m in _adapters.items() if m}
    for a in [a for a in adapters if enabled_adapters.get(a.category) == a.name]:
        a.enabled = True
    registry = adapter_registry.get()
    for a in adapters:
        module = ".".join(a.module.split(".")[-2:])
        remove = next(
            (
                _a
                for _a in registry
                if ".".join(_a.module.split(".")[-2:]) == module and _a.pkg == "acb"
            ),
            None,
        )
        if remove:
            registry.remove(remove)
        if module in [".".join(_a.module.split(".")[-2:]) for _a in registry]:
            continue
        registry.append(a)
    return adapters



from acb.config import AdapterBase, Settings


class ModelsBaseSettings(Settings): ...


class ModelsBase(AdapterBase):
    class SqlModels: ...

    sql = SqlModels()

    class NosqlModels: ...

    nosql = NosqlModels()



import typing as t
from gzip import GzipFile
from io import BytesIO
from pathlib import Path

import brotli
from pydantic import BaseModel


class Compress(BaseModel):
    @staticmethod
    def gzip(
        content: str | bytes,
        path: t.Optional[str | Path] = None,
        compresslevel: int = 6,
    ) -> bytes:
        gzip_buffer = BytesIO()
        gz = GzipFile(path, "wb", compresslevel, gzip_buffer)
        if path:
            if isinstance(content, bytes):
                gz.write(content)
            else:
                gz.write(content.encode())
        gz.close()
        return gzip_buffer.getvalue()

    @staticmethod
    def brotli(data: bytes | str, level: int = 3) -> bytes:
        if isinstance(data, str):
            data = data.encode()
        return brotli.compress(data, quality=level)


compress = Compress()


class Decompress(BaseModel):
    @staticmethod
    def brotli(data: bytes) -> str:
        return brotli.decompress(data)

    @staticmethod
    def gzip(content: t.Any) -> str:
        gzip_buffer = BytesIO(content)
        return GzipFile(None, "rb", fileobj=gzip_buffer).read().decode()


decompress = Decompress()



import typing as t
from abc import ABC, abstractmethod

from pydantic import BaseModel
from acb.config import AdapterBase, Settings


class DnsRecord(BaseModel):
    name: t.Optional[str] = None
    type: str = "TXT"
    ttl: int = 300
    rrdata: t.Optional[str | list[t.Any]] = None


class DnsBaseSettings(Settings): ...


class DnsBase(AdapterBase, ABC):
    client: t.Optional[t.Any] = None
    zone: t.Optional[t.Any] = None

    @abstractmethod
    def create_zone(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def list_records(self) -> list[DnsRecord]:
        raise NotImplementedError

    @abstractmethod
    async def create_records(self, records: list[DnsRecord] | DnsRecord) -> None:
        raise NotImplementedError
